# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Ecosystem helper configurations."""
import typing

from ..cache import Cache

timeout = 30  # Timeout for HTTP(S) requests

# Used for checking out git repositories
# Intended to be set in worker.py
work_dir: typing.Optional[str] = None

shared_cache: typing.Optional[Cache] = None


def set_cache(cache: typing.Optional[Cache]):
  """Configures and enables the redis caching layer"""
  global shared_cache
  shared_cache = cache
