// Code generated by scripts/generate-go-constants.py. DO NOT EDIT.
package osvschema

const SchemaVersion = "1.7.3"

type CreditType string

const (
	CreditAnalyst              CreditType = "ANALYST"
	CreditCoordinator          CreditType = "COORDINATOR"
	CreditFinder               CreditType = "FINDER"
	CreditOther                CreditType = "OTHER"
	CreditRemediationDeveloper CreditType = "REMEDIATION_DEVELOPER"
	CreditRemediationReviewer  CreditType = "REMEDIATION_REVIEWER"
	CreditRemediationVerifier  CreditType = "REMEDIATION_VERIFIER"
	CreditReporter             CreditType = "REPORTER"
	CreditSponsor              CreditType = "SPONSOR"
	CreditTool                 CreditType = "TOOL"
)

type Ecosystem string

const (
	EcosystemAlmaLinux                  Ecosystem = "AlmaLinux"
	EcosystemAlpaquita                  Ecosystem = "Alpaquita"
	EcosystemAlpine                     Ecosystem = "Alpine"
	EcosystemAndroid                    Ecosystem = "Android"
	EcosystemBellSoftHardenedContainers Ecosystem = "BellSoft Hardened Containers"
	EcosystemBioconductor               Ecosystem = "Bioconductor"
	EcosystemBitnami                    Ecosystem = "Bitnami"
	EcosystemChainguard                 Ecosystem = "Chainguard"
	EcosystemConanCenter                Ecosystem = "ConanCenter"
	EcosystemCRAN                       Ecosystem = "CRAN"
	EcosystemCratesIO                   Ecosystem = "crates.io"
	EcosystemDebian                     Ecosystem = "Debian"
	EcosystemEcho                       Ecosystem = "Echo"
	EcosystemGHC                        Ecosystem = "GHC"
	EcosystemGitHubActions              Ecosystem = "GitHub Actions"
	EcosystemGo                         Ecosystem = "Go"
	EcosystemHackage                    Ecosystem = "Hackage"
	EcosystemHex                        Ecosystem = "Hex"
	EcosystemKubernetes                 Ecosystem = "Kubernetes"
	EcosystemLinux                      Ecosystem = "Linux"
	EcosystemMageia                     Ecosystem = "Mageia"
	EcosystemMaven                      Ecosystem = "Maven"
	EcosystemMinimOS                    Ecosystem = "MinimOS"
	EcosystemNPM                        Ecosystem = "npm"
	EcosystemNuGet                      Ecosystem = "NuGet"
	EcosystemOpenEuler                  Ecosystem = "openEuler"
	EcosystemOpenSUSE                   Ecosystem = "openSUSE"
	EcosystemOSSFuzz                    Ecosystem = "OSS-Fuzz"
	EcosystemPackagist                  Ecosystem = "Packagist"
	EcosystemPhotonOS                   Ecosystem = "Photon OS"
	EcosystemPub                        Ecosystem = "Pub"
	EcosystemPyPI                       Ecosystem = "PyPI"
	EcosystemRedHat                     Ecosystem = "Red Hat"
	EcosystemRockyLinux                 Ecosystem = "Rocky Linux"
	EcosystemRubyGems                   Ecosystem = "RubyGems"
	EcosystemSUSE                       Ecosystem = "SUSE"
	EcosystemSwiftURL                   Ecosystem = "SwiftURL"
	EcosystemUbuntu                     Ecosystem = "Ubuntu"
	EcosystemWolfi                      Ecosystem = "Wolfi"
)

type RangeType string

const (
	RangeEcosystem RangeType = "ECOSYSTEM"
	RangeGit       RangeType = "GIT"
	RangeSemver    RangeType = "SEMVER"
)

type ReferenceType string

const (
	ReferenceAdvisory   ReferenceType = "ADVISORY"
	ReferenceArticle    ReferenceType = "ARTICLE"
	ReferenceDetection  ReferenceType = "DETECTION"
	ReferenceDiscussion ReferenceType = "DISCUSSION"
	ReferenceEvidence   ReferenceType = "EVIDENCE"
	ReferenceFix        ReferenceType = "FIX"
	ReferenceGit        ReferenceType = "GIT"
	ReferenceIntroduced ReferenceType = "INTRODUCED"
	ReferencePackage    ReferenceType = "PACKAGE"
	ReferenceReport     ReferenceType = "REPORT"
	ReferenceWeb        ReferenceType = "WEB"
)

type SeverityType string

const (
	SeverityCvssV2 SeverityType = "CVSS_V2"
	SeverityCvssV3 SeverityType = "CVSS_V3"
	SeverityCvssV4 SeverityType = "CVSS_V4"
	SeverityUbuntu SeverityType = "Ubuntu"
)
